////////////////////////////////////////////////////////////////////////////////
// AReflection Library
// Copyright (c) 2004 by Arne Adams
// Permission to use, copy, modify, distribute and sell this software for any 
//     purpose is hereby granted without fee, provided that the above copyright 
//     notice appear in all copies and that both that copyright notice and this 
//     permission notice appear in supporting documentation.
// The author makes no representations about the suitability of this software 
//     for any purpose. It is provided "as is" without express or implied 
//     warranty.
////////////////////////////////////////////////////////////////////////////////
#ifndef REFLIB_FLATFOREACH_H_INC_
#define REFLIB_FLATFOREACH_H_INC_

#include <typeinfo>

#include "../baseclassiter.h"

#include "../priviligedaccess.h"

#include "../conditional_lock.h"

namespace refl
{
    template<class ReflectedClass, class MemberKinds> struct AllAttributes
    {
        enum {TotalNumAttributes = ReflectedClass::NumAttributes};
        template<int Pos> struct SingleAttributeHelper
        {
            template<class Function> static void
            processSingleAttribute(ReflectedClass& recordData,
                                    Function& function)
            {  
                // inheriting like in the reference implementation gives another
                // ICE - note that I cannot even inherit SingleAttributeHelper 
                // from this class:
                // template<class ReflectedClass, class MemberKinds, int Pos, 
                //     class HoldsImplForMatchingKind> 
                // struct PleaseDontICE{};
                typedef MatchingAttributeHandler<ReflectedClass,MemberKinds,Pos,
                                                SingleAttributeHelper<Pos> >
                    ICEWorkaround;
                ICEWorkaround::processSingleAttribute(recordData,function);
            } 
            template<class Function, class ValueType> static void 
            processMatchingAttribute(ReflectedClass& recordData,
                                    Function& function,
                                    ValueType& attribute)
            {  
                // MSVC++ 6.0 does not allow these typedefs at class scope.
                // Hence accessing a these types is implemented per copy and 
                // paste of these lines
                typedef typename ReflectedClass::template 
                    GetNameTag<Pos,ReflectedClass::ThisClass>::type 
                    CurrentAttributeTag;
#ifdef _DEBUG
                const char* AttributeName = typeid(CurrentAttributeTag).name();
#endif
                function(attribute,
                        CurrentAttributeTag(),PassScope<ReflectedClass>(&recordData));   
                // This helps for nested attributes - especially to pass the 
            }   // type of the enclosing class back to the caller
            // overloads to be able to access classvariables without having
            // an instance of ReflectedClass
            template<class Function> static void
            processSingleAttribute(Function& function)
            {  
                // inheriting like in the reference implementation gives another
                // ICE - note that I cannot even inherit SingleAttributeHelper 
                // from this class:
                // template<class ReflectedClass, class MemberKinds, int Pos, 
                //     class HoldsImplForMatchingKind> 
                // struct PleaseDontICE{};
                typedef MatchingAttributeHandler<ReflectedClass,MemberKinds,Pos,
                                                SingleAttributeHelper<Pos> >
                    ICEWorkaround;
                ICEWorkaround::processSingleAttribute(function);
            } 
            template<class Function, class ValueType> static void 
            processMatchingAttribute(Function& function,
                                    ValueType& attribute)
            {  
                // MSVC++ 6.0 does not allow these typedefs at class scope.
                // Hence accessing a these types is implemented per copy and 
                // paste of these lines
                typedef typename ReflectedClass::template 
                    GetNameTag<Pos,ReflectedClass::ThisClass>::type 
                    CurrentAttributeTag;
#ifdef _DEBUG
                const char* AttributeName = typeid(CurrentAttributeTag).name();
#endif
                function(attribute,
                        CurrentAttributeTag(),PassType<ReflectedClass>());
                // This helps for nested attributes - especially to pass the 
            }   // type of the enclosing class back to the caller
        };
        //template<int Pos> struct Iter: SingleAttributeHelper<Pos> 
        template<int RestToDo> struct Iter: 
                        SingleAttributeHelper<TotalNumAttributes - RestToDo> 
        {
            typedef SingleAttributeHelper<TotalNumAttributes - RestToDo> Inherited;
            template<class Function> static void 
            process(ReflectedClass& recordData,Function& function)
            {
                Inherited::processSingleAttribute(recordData,function);
                Iter<RestToDo-1>::process(recordData,function);
            }
            template<class Function> static void 
            process(Function& function)
            {
                Inherited::processSingleAttribute(function);
                Iter<RestToDo-1>::process(function);
            }
        };
        // this is illegal in standard c++ : full specialization definition
        // inside the body of a primary template 
        //template<> struct Iter<TotalNumAttributes - 1, Dummy>: 
        //                        SingleAttributeHelper<TotalNumAttributes - 1>
        template<> struct Iter<1>: 
                                SingleAttributeHelper<TotalNumAttributes - 1>
        {
            typedef SingleAttributeHelper<TotalNumAttributes - 1> Inherited;
            template<class Function> static void 
            process(ReflectedClass& recordData,Function& function)
            {
                Inherited::processSingleAttribute(recordData,function);
            }
            template<class Function> static void 
            process(Function& function)
            {
                Inherited::processSingleAttribute(function);
            }
        };
        // having support for baseclasses changes in the reflected members could
        // otherwise "silently" change the compilability of derived classes.
        //template<> struct Iter<TotalNumAttributes>
        template<> struct Iter<0>
        {
            template<class Function> static void 
            process(ReflectedClass& ,Function& )
            {}
            template<class Function> static void process(Function& )
            {}
        };

        template<class Function> static void 
        doForeach(ReflectedClass& dataTuple, Function& function)
        {
            ConditionalLock<ReflectedClass> lock(dataTuple);
            Iter<TotalNumAttributes>::process(dataTuple,function);
        }
        template<class Function> static void 
        doForeach(Function& function)
        {
            ConditionalLock<ReflectedClass> lock;
            Iter<TotalNumAttributes>::process(function);
        }
    };
    
    // All usages so far can take advantage of the AttributeTag supplied - hence
    // this Tag will be passed as Parameter to the function back
    template<class ReflectedClass, class Function, class MemberKinds> 
    void foreachAttribute(ReflectedClass& dataTuple, 
                        Function& function,
                        MemberKinds)   
    {
        AllAttributes<ReflectedClass,MemberKinds>::doForeach(dataTuple,
                                                            function);
    }
    // without a ReflectedClass instance only static members can be iterated
    template<class ReflectedClass, class Function> 
    void foreachAttribute(Function& function, 
                    ReflectedClass* = NULL)
    {
        AllAttributes<ReflectedClass,ClassVariableTag>::doForeach(function);
    }

    template<class Function,class MemberKinds> struct ForeachAttribute;
    // some algorithm may explicitly need to traverse attributes of the 
    // baseclasses others don't (for instance if initReflectedMembers
    // is called in each ctor of each reflected class then this function
    // does has to use foreachAttribute as its back-end (to avoid unnecessary
    // assignments).
    template<class ReflectedClass, class Function, class MemberKinds> 
    void foreachAttributeInclBaseClasses(ReflectedClass& dataTuple, 
                                        Function& function,
                                        MemberKinds)   
    {
        ForeachAttribute<Function,MemberKinds> directMembersIteration(function);
        foreachBaseClass(dataTuple,directMembersIteration);
        directMembersIteration(dataTuple);
    }
    // ReflectedClass is needed as part of the function type because otherwise
    // MSVC++ 6.0 is not able to link instantiations that differ in 
    // ReflectedClass correctly
    template<class ReflectedClass, class Function> 
    void foreachAttributeInclBaseClasses(Function& function, ReflectedClass* = NULL)
    {
        ForeachAttribute<Function,ClassVariableTag> 
                                            directMembersIteration(function);
        foreachBaseClass<ReflectedClass>(directMembersIteration);
        directMembersIteration(PassType<ReflectedClass>(), ClassVariableTag());
    }
    // having this as a function object facilitates its use as parameter
    // to other compile-time recursions.
    template<class Function, class MemberKinds> struct ForeachAttribute
    {
        ForeachAttribute(Function& function):function_(function)
        {}
        template<class ReflectedClass> void 
        operator()(ReflectedClass& dataTuple)const
        {
            foreachAttribute(dataTuple,function_,MemberKinds());
        }
        // ClassVariableTag is added here because VC++ 6.0 has no partial 
        // ordering of function templates
        template<class ReflectedClass> void operator()(PassType<ReflectedClass>,
                                                        ClassVariableTag)const
        {
            foreachAttribute<ReflectedClass>(function_);
        }
        Function& function_;
    };
}

#endif
